@@ -17,6 +17,7 @@ __all__ = ['AMQPDriverBase']
 
 import logging
 import threading
+import time
 import uuid
 
 import cachetools
@@ -27,58 +28,46 @@ from oslo_messaging._drivers import amqp as rpc_amqp
 from oslo_messaging._drivers import base
 from oslo_messaging._drivers import common as rpc_common
 from oslo_messaging._i18n import _
+from oslo_messaging._i18n import _LE
 from oslo_messaging._i18n import _LI
 from oslo_messaging._i18n import _LW
 
 LOG = logging.getLogger(__name__)
 
 
-class AMQPIncomingMessage(base.IncomingMessage):
+class AMQPIncomingMessage(base.RpcIncomingMessage):
 
     def __init__(self, listener, ctxt, message, unique_id, msg_id, reply_q,
                  obsolete_reply_queues):
-        super(AMQPIncomingMessage, self).__init__(listener, ctxt,
-                                                  dict(message))
+        super(AMQPIncomingMessage, self).__init__(ctxt, message)
+        self.listener = listener
 
         self.unique_id = unique_id
         self.msg_id = msg_id
         self.reply_q = reply_q
-        self.acknowledge_callback = message.acknowledge
-        self.requeue_callback = message.requeue
         self._obsolete_reply_queues = obsolete_reply_queues
 
-    def _send_reply(self, conn, reply=None, failure=None,
-                    ending=False, log_failure=True):
-        if (self.reply_q and
-            not self._obsolete_reply_queues.reply_q_valid(self.reply_q,
-                                                          self.msg_id)):
+    def _send_reply(self, conn, reply=None, failure=None, log_failure=True):
+        if not self._obsolete_reply_queues.reply_q_valid(self.reply_q,
+                                                         self.msg_id):
             return
 
         if failure:
             failure = rpc_common.serialize_remote_exception(failure,
                                                             log_failure)
-
-        msg = {'result': reply, 'failure': failure}
-        if ending:
-            msg['ending'] = True
-
+        # NOTE(sileht): ending can be removed in N*, see Listener.wait()
+        # for more detail.
+        msg = {'result': reply, 'failure': failure, 'ending': True,
+               '_msg_id': self.msg_id}
         rpc_amqp._add_unique_id(msg)
+        unique_id = msg[rpc_amqp.UNIQUE_ID]
 
-        # If a reply_q exists, add the msg_id to the reply and pass the
-        # reply_q to direct_send() to use it as the response queue.
-        # Otherwise use the msg_id for backward compatibility.
-        if self.reply_q:
-            msg['_msg_id'] = self.msg_id
-            try:
-                conn.direct_send(self.reply_q, rpc_common.serialize_msg(msg))
-            except rpc_amqp.AMQPDestinationNotFound:
-                self._obsolete_reply_queues.add(self.reply_q, self.msg_id)
-        else:
-            # TODO(sileht): look at which version of oslo-incubator rpc
-            # send need this, but I guess this is older than icehouse
-            # if this is icehouse, we can drop this at M
-            # if this is havana, we can drop this now.
-            conn.direct_send(self.msg_id, rpc_common.serialize_msg(msg))
+        LOG.debug("sending reply msg_id: %(msg_id)s "
+                  "reply queue: %(reply_q)s", {
+                      'msg_id': self.msg_id,
+                      'unique_id': unique_id,
+                      'reply_q': self.reply_q})
+        conn.direct_send(self.reply_q, rpc_common.serialize_msg(msg))
 
     def reply(self, reply=None, failure=None, log_failure=True):
         if not self.msg_id:
@@ -87,23 +76,45 @@ class AMQPIncomingMessage(base.IncomingMessage):
             return
 
         # NOTE(sileht): return without hold the a connection if possible
-        if (self.reply_q and
-            not self._obsolete_reply_queues.reply_q_valid(self.reply_q,
-                                                          self.msg_id)):
+        if not self._obsolete_reply_queues.reply_q_valid(self.reply_q,
+                                                         self.msg_id):
             return
 
-        with self.listener.driver._get_connection(
-                rpc_amqp.PURPOSE_SEND) as conn:
-            if self.listener.driver.send_single_reply:
-                self._send_reply(conn, reply, failure, log_failure=log_failure,
-                                 ending=True)
-            else:
-                self._send_reply(conn, reply, failure, log_failure=log_failure)
-                self._send_reply(conn, ending=True)
+        # NOTE(sileht): we read the configuration value from the driver
+        # to be able to backport this change in previous version that
+        # still have the qpid driver
+        duration = self.listener.driver.missing_destination_retry_timeout
+        timer = rpc_common.DecayingTimer(duration=duration)
+        timer.start()
+
+        while True:
+            try:
+                with self.listener.driver._get_connection(
+                        rpc_common.PURPOSE_SEND) as conn:
+                    self._send_reply(conn, reply, failure,
+                                     log_failure=log_failure)
+                return
+            except rpc_amqp.AMQPDestinationNotFound:
+                if timer.check_return() > 0:
+                    LOG.debug(("The reply %(msg_id)s cannot be sent  "
+                               "%(reply_q)s reply queue don't exist, "
+                               "retrying..."), {
+                                   'msg_id': self.msg_id,
+                                   'reply_q': self.reply_q})
+                    time.sleep(0.25)
+                else:
+                    self._obsolete_reply_queues.add(self.reply_q, self.msg_id)
+                    LOG.info(_LI("The reply %(msg_id)s cannot be sent  "
+                                 "%(reply_q)s reply queue don't exist after "
+                                 "%(duration)s sec abandoning..."), {
+                                     'msg_id': self.msg_id,
+                                     'reply_q': self.reply_q,
+                                     'duration': duration})
+                    return
 
     def acknowledge(self):
+        self.message.acknowledge()
         self.listener.msg_id_cache.add(self.unique_id)
-        self.acknowledge_callback()
 
     def requeue(self):
         # NOTE(sileht): In case of the connection is lost between receiving the
@@ -112,7 +123,7 @@ class AMQPIncomingMessage(base.IncomingMessage):
         # msg_id_cache, the message will be reconsumed, the only difference is
         # the message stay at the beginning of the queue instead of moving to
         # the end.
-        self.requeue_callback()
+        self.message.requeue()
 
 
 class ObsoleteReplyQueuesCache(object):
@@ -155,15 +166,16 @@ class ObsoleteReplyQueuesCache(object):
         self._no_reply_log(reply_q, msg_id)
 
     def _no_reply_log(self, reply_q, msg_id):
-        LOG.warn(_LW("%(reply_queue)s doesn't exists, drop reply to "
-                     "%(msg_id)s"), {'reply_queue': reply_q,
-                                     'msg_id': msg_id})
+        LOG.warning(_LW("%(reply_queue)s doesn't exists, drop reply to "
+                        "%(msg_id)s"), {'reply_queue': reply_q,
+                                        'msg_id': msg_id})
 
 
 class AMQPListener(base.Listener):
 
     def __init__(self, driver, conn):
-        super(AMQPListener, self).__init__(driver)
+        super(AMQPListener, self).__init__(driver.prefetch_size)
+        self.driver = driver
         self.conn = conn
         self.msg_id_cache = rpc_amqp._MsgIdCache()
         self.incoming = []
@@ -171,7 +183,7 @@ class AMQPListener(base.Listener):
         self._obsolete_reply_queues = ObsoleteReplyQueuesCache()
 
     def __call__(self, message):
-        ctxt = rpc_amqp.unpack_context(self.conf, message)
+        ctxt = rpc_amqp.unpack_context(message)
 
         # FIXME(sileht): Don't log the message until strutils is more
         # efficient, (rpc_amqp.unpack_context already log the context)
@@ -180,6 +192,9 @@ class AMQPListener(base.Listener):
 
         unique_id = self.msg_id_cache.check_duplicate_message(message)
 
+        LOG.debug("received message msg_id: %(msg_id)s reply to %(queue)s", {
+            'queue': ctxt.reply_q, 'msg_id': ctxt.msg_id})
+
         self.incoming.append(AMQPIncomingMessage(self,
                                                  ctxt.to_dict(),
                                                  message,
@@ -188,6 +203,7 @@ class AMQPListener(base.Listener):
                                                  ctxt.reply_q,
                                                  self._obsolete_reply_queues))
 
+    @base.batch_poll_helper
     def poll(self, timeout=None):
         while not self._stopped.is_set():
             if self.incoming:
@@ -234,10 +250,11 @@ class ReplyWaiters(object):
     def add(self, msg_id):
         self._queues[msg_id] = moves.queue.Queue()
         if len(self._queues) > self._wrn_threshold:
-            LOG.warn('Number of call queues is greater than warning '
-                     'threshold: %d. There could be a leak. Increasing'
-                     ' threshold to: %d', self._wrn_threshold,
-                     self._wrn_threshold * 2)
+            LOG.warning(_LW('Number of call queues is greater than warning '
+                            'threshold: %(old_threshold)s. There could be a '
+                            'leak. Increasing threshold to: %(threshold)s'),
+                        {'old_threshold': self._wrn_threshold,
+                         'threshold': self._wrn_threshold * 2})
             self._wrn_threshold *= 2
 
     def remove(self, msg_id):
@@ -270,12 +287,14 @@ class ReplyWaiter(object):
             try:
                 self.conn.consume()
             except Exception:
-                LOG.exception("Failed to process incoming message, "
-                              "retrying...")
+                LOG.exception(_LE("Failed to process incoming message, "
+                              "retrying..."))
 
     def __call__(self, message):
         message.acknowledge()
         incoming_msg_id = message.pop('_msg_id', None)
+        if message.get('ending'):
+            LOG.debug("received reply msg_id: %s", incoming_msg_id)
         self.waiters.put(incoming_msg_id, message)
 
     def listen(self, msg_id):
@@ -329,10 +348,10 @@ class ReplyWaiter(object):
 
 
 class AMQPDriverBase(base.BaseDriver):
+    missing_destination_retry_timeout = 0
 
     def __init__(self, conf, url, connection_pool,
-                 default_exchange=None, allowed_remote_exmods=None,
-                 send_single_reply=False):
+                 default_exchange=None, allowed_remote_exmods=None):
         super(AMQPDriverBase, self).__init__(conf, url, default_exchange,
                                              allowed_remote_exmods)
 
@@ -345,14 +364,12 @@ class AMQPDriverBase(base.BaseDriver):
         self._reply_q_conn = None
         self._waiter = None
 
-        self.send_single_reply = send_single_reply
-
     def _get_exchange(self, target):
         return target.exchange or self._default_exchange
 
-    def _get_connection(self, purpose=rpc_amqp.PURPOSE_SEND):
-        return rpc_amqp.ConnectionContext(self._connection_pool,
-                                          purpose=purpose)
+    def _get_connection(self, purpose=rpc_common.PURPOSE_SEND):
+        return rpc_common.ConnectionContext(self._connection_pool,
+                                            purpose=purpose)
 
     def _get_reply_q(self):
         with self._reply_q_lock:
@@ -361,7 +378,7 @@ class AMQPDriverBase(base.BaseDriver):
 
             reply_q = 'reply_' + uuid.uuid4().hex
 
-            conn = self._get_connection(rpc_amqp.PURPOSE_LISTEN)
+            conn = self._get_connection(rpc_common.PURPOSE_LISTEN)
 
             self._waiter = ReplyWaiter(reply_q, conn,
                                        self._allowed_remote_exmods)
@@ -389,10 +406,11 @@ class AMQPDriverBase(base.BaseDriver):
         if wait_for_reply:
             msg_id = uuid.uuid4().hex
             msg.update({'_msg_id': msg_id})
-            LOG.debug('MSG_ID is %s', msg_id)
             msg.update({'_reply_q': self._get_reply_q()})
 
         rpc_amqp._add_unique_id(msg)
+        unique_id = msg[rpc_amqp.UNIQUE_ID]
+
         rpc_amqp.pack_context(msg, context)
 
         if envelope:
@@ -400,21 +418,37 @@ class AMQPDriverBase(base.BaseDriver):
 
         if wait_for_reply:
             self._waiter.listen(msg_id)
+            log_msg = "CALL msg_id: %s " % msg_id
+        else:
+            log_msg = "CAST unique_id: %s " % unique_id
 
         try:
-            with self._get_connection(rpc_amqp.PURPOSE_SEND) as conn:
+            with self._get_connection(rpc_common.PURPOSE_SEND) as conn:
                 if notify:
-                    conn.notify_send(self._get_exchange(target),
-                                     target.topic, msg, retry=retry)
+                    exchange = self._get_exchange(target)
+                    log_msg += "NOTIFY exchange '%(exchange)s'" \
+                               " topic '%(topic)s'" % {
+                                   'exchange': exchange,
+                                   'topic': target.topic}
+                    LOG.debug(log_msg)
+                    conn.notify_send(exchange, target.topic, msg, retry=retry)
                 elif target.fanout:
+                    log_msg += "FANOUT topic '%(topic)s'" % {
+                        'topic': target.topic}
+                    LOG.debug(log_msg)
                     conn.fanout_send(target.topic, msg, retry=retry)
                 else:
                     topic = target.topic
+                    exchange = self._get_exchange(target)
                     if target.server:
                         topic = '%s.%s' % (target.topic, target.server)
-                    conn.topic_send(exchange_name=self._get_exchange(target),
-                                    topic=topic, msg=msg, timeout=timeout,
-                                    retry=retry)
+                    log_msg += "exchange '%(exchange)s'" \
+                               " topic '%(topic)s'" % {
+                                   'exchange': exchange,
+                                   'topic': target.topic}
+                    LOG.debug(log_msg)
+                    conn.topic_send(exchange_name=exchange, topic=topic,
+                                    msg=msg, timeout=timeout, retry=retry)
 
             if wait_for_reply:
                 result = self._waiter.wait(msg_id, timeout)
@@ -435,7 +469,7 @@ class AMQPDriverBase(base.BaseDriver):
                           envelope=(version == 2.0), notify=True, retry=retry)
 
     def listen(self, target):
-        conn = self._get_connection(rpc_amqp.PURPOSE_LISTEN)
+        conn = self._get_connection(rpc_common.PURPOSE_LISTEN)
 
         listener = AMQPListener(self, conn)
 
@@ -451,7 +485,7 @@ class AMQPDriverBase(base.BaseDriver):
         return listener
 
     def listen_for_notifications(self, targets_and_priorities, pool):
-        conn = self._get_connection(rpc_amqp.PURPOSE_LISTEN)
+        conn = self._get_connection(rpc_common.PURPOSE_LISTEN)
 
         listener = AMQPListener(self, conn)
         for target, priority in targets_and_priorities:
