@@ -22,14 +22,11 @@ import itertools
 
 from oslo_messaging._drivers import amqp
 from oslo_messaging._drivers import base as drivers_base
-from oslo_messaging._drivers import impl_qpid
 from oslo_messaging._drivers import impl_rabbit
 from oslo_messaging._drivers import impl_zmq
-from oslo_messaging._drivers import matchmaker
-from oslo_messaging._drivers import matchmaker_redis
-from oslo_messaging._drivers import matchmaker_ring
 from oslo_messaging._drivers.protocols.amqp import opts as amqp_opts
-from oslo_messaging._executors import base
+from oslo_messaging._drivers.zmq_driver.matchmaker import matchmaker_redis
+from oslo_messaging._executors import impl_pooledexecutor
 from oslo_messaging.notify import notifier
 from oslo_messaging.rpc import client
 from oslo_messaging import transport
@@ -37,9 +34,7 @@ from oslo_messaging import transport
 _global_opt_lists = [
     drivers_base.base_opts,
     impl_zmq.zmq_opts,
-    matchmaker.matchmaker_opts,
-    base._pool_opts,
-    notifier._notifier_opts,
+    impl_pooledexecutor._pool_opts,
     client._client_opts,
     transport._transport_opts,
 ]
@@ -47,12 +42,10 @@ _global_opt_lists = [
 _opts = [
     (None, list(itertools.chain(*_global_opt_lists))),
     ('matchmaker_redis', matchmaker_redis.matchmaker_redis_opts),
-    ('matchmaker_ring', matchmaker_ring.matchmaker_opts),
     ('oslo_messaging_amqp', amqp_opts.amqp1_opts),
+    ('oslo_messaging_notifications', notifier._notifier_opts),
     ('oslo_messaging_rabbit', list(itertools.chain(amqp.amqp_opts,
                                                    impl_rabbit.rabbit_opts))),
-    ('oslo_messaging_qpid', list(itertools.chain(amqp.amqp_opts,
-                                                 impl_qpid.qpid_opts)))
 ]
 
 
@@ -76,3 +69,23 @@ def list_opts():
     :returns: a list of (group_name, opts) tuples
     """
     return [(g, copy.deepcopy(o)) for g, o in _opts]
+
+
+def set_defaults(conf, executor_thread_pool_size=None):
+    """Set defaults for configuration variables.
+
+    Overrides default options values.
+
+    :param conf: Config instance specified to set default options in it. Using
+     of instances instead of a global config object prevents conflicts between
+     options declaration.
+    :type conf: oslo.config.cfg.ConfigOpts instance.
+
+    :keyword executor_thread_pool_size: Size of executor thread pool.
+    :type executor_thread_pool_size: int
+    :default executor_thread_pool_size: None
+
+    """
+    if executor_thread_pool_size is not None:
+        conf.set_default('executor_thread_pool_size',
+                         executor_thread_pool_size)
