@@ -14,8 +14,8 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import contextlib
 import threading
+import time
 
 # eventlet 0.16 with monkey patching does not work yet on Python 3,
 # so make aioeventlet, eventlet and trollius import optional
@@ -28,7 +28,6 @@ try:
 except ImportError:
     eventlet = None
 import testscenarios
-import testtools
 try:
     import trollius
 except ImportError:
@@ -45,6 +44,7 @@ try:
 except ImportError:
     impl_eventlet = None
 from oslo_messaging._executors import impl_thread
+from oslo_messaging import dispatcher as dispatcher_base
 from oslo_messaging.tests import utils as test_utils
 from six.moves import mock
 
@@ -72,15 +72,21 @@ class TestExecutor(test_utils.BaseTestCase):
         thread = threading.Thread(target=target, args=(executor,))
         thread.daemon = True
         thread.start()
-        thread.join(timeout=30)
+        return thread
 
-    def test_executor_dispatch(self):
+    def _create_dispatcher(self):
         if impl_aioeventlet is not None:
             aioeventlet_class = impl_aioeventlet.AsyncioEventletExecutor
         else:
             aioeventlet_class = None
         is_aioeventlet = (self.executor == aioeventlet_class)
 
+        if impl_blocking is not None:
+            blocking_class = impl_blocking.BlockingExecutor
+        else:
+            blocking_class = None
+        is_blocking = (self.executor == blocking_class)
+
         if is_aioeventlet:
             policy = aioeventlet.EventLoopPolicy()
             trollius.set_event_loop_policy(policy)
@@ -106,121 +112,99 @@ class TestExecutor(test_utils.BaseTestCase):
 
             @trollius.coroutine
             def simple_coroutine(value):
-                yield None
                 raise trollius.Return(value)
 
             endpoint = mock.MagicMock(return_value=simple_coroutine('result'))
             event = eventlet.event.Event()
+        elif is_blocking:
+            def run_executor(executor):
+                executor.start()
+                executor.execute()
+                executor.wait()
+
+            endpoint = mock.MagicMock(return_value='result')
+            event = None
         else:
             def run_executor(executor):
                 executor.start()
                 executor.wait()
 
             endpoint = mock.MagicMock(return_value='result')
+            event = None
 
-        class Dispatcher(object):
+        class Dispatcher(dispatcher_base.DispatcherBase):
             def __init__(self, endpoint):
                 self.endpoint = endpoint
                 self.result = "not set"
 
-            @contextlib.contextmanager
-            def __call__(self, incoming, executor_callback=None):
-                if executor_callback is not None:
-                    def callback():
-                        result = executor_callback(self.endpoint,
-                                                   incoming.ctxt,
-                                                   incoming.message)
-                        self.result = result
-                        return result
-                    yield callback
-                    event.send()
+            def _listen(self, transport):
+                pass
+
+            def callback(self, incoming, executor_callback):
+                if executor_callback is None:
+                    result = self.endpoint(incoming.ctxt,
+                                           incoming.message)
                 else:
-                    def callback():
-                        result = self.endpoint(incoming.ctxt, incoming.message)
-                        self.result = result
-                        return result
-                    yield callback
+                    result = executor_callback(self.endpoint,
+                                               incoming.ctxt,
+                                               incoming.message)
+                if is_aioeventlet:
+                    event.send()
+                self.result = result
+                return result
 
-        listener = mock.Mock(spec=['poll'])
-        dispatcher = Dispatcher(endpoint)
+            def __call__(self, incoming, executor_callback=None):
+                return dispatcher_base.DispatcherExecutorContext(
+                    incoming[0], self.callback, executor_callback)
+
+        return Dispatcher(endpoint), endpoint, event, run_executor
+
+    def test_slow_wait(self):
+        dispatcher, endpoint, event, run_executor = self._create_dispatcher()
+        listener = mock.Mock(spec=['poll', 'stop'])
         executor = self.executor(self.conf, listener, dispatcher)
+        incoming_message = mock.MagicMock(ctxt={}, message={'payload': 'data'})
 
-        incoming_message = mock.MagicMock(ctxt={},
-                                          message={'payload': 'data'})
-
-        def fake_poll(timeout=None):
-            if is_aioeventlet:
-                if listener.poll.call_count == 1:
-                    return incoming_message
-                event.wait()
+        def fake_poll(timeout=None, prefetch_size=1):
+            time.sleep(0.1)
+            if listener.poll.call_count == 10:
+                if event is not None:
+                    event.wait()
                 executor.stop()
             else:
-                if listener.poll.call_count == 1:
-                    return incoming_message
-                executor.stop()
+                return incoming_message
 
         listener.poll.side_effect = fake_poll
+        thread = self._run_in_thread(run_executor, executor)
+        self.assertFalse(executor.wait(timeout=0.1))
+        thread.join()
+        self.assertTrue(executor.wait())
 
-        self._run_in_thread(run_executor, executor)
+    def test_dead_wait(self):
+        (dispatcher, _endpoint, _event,
+            _run_executor) = self._create_dispatcher()
+        listener = mock.Mock(spec=['poll', 'stop'])
+        executor = self.executor(self.conf, listener, dispatcher)
+        executor.stop()
+        self.assertTrue(executor.wait())
 
+    def test_executor_dispatch(self):
+        dispatcher, endpoint, event, run_executor = self._create_dispatcher()
+        listener = mock.Mock(spec=['poll', 'stop'])
+        executor = self.executor(self.conf, listener, dispatcher)
+        incoming_message = mock.MagicMock(ctxt={}, message={'payload': 'data'})
+
+        def fake_poll(timeout=None, prefetch_size=1):
+            if listener.poll.call_count == 1:
+                return [incoming_message]
+            if event is not None:
+                event.wait()
+            executor.stop()
+
+        listener.poll.side_effect = fake_poll
+        thread = self._run_in_thread(run_executor, executor)
+        thread.join()
         endpoint.assert_called_once_with({}, {'payload': 'data'})
         self.assertEqual(dispatcher.result, 'result')
 
 TestExecutor.generate_scenarios()
-
-
-class ExceptedException(Exception):
-    pass
-
-
-class EventletContextManagerSpawnTest(test_utils.BaseTestCase):
-    @testtools.skipIf(impl_eventlet is None, "Eventlet not available")
-    def setUp(self):
-        super(EventletContextManagerSpawnTest, self).setUp()
-        self.before = mock.Mock()
-        self.callback = mock.Mock()
-        self.after = mock.Mock()
-        self.exception_call = mock.Mock()
-
-        @contextlib.contextmanager
-        def context_mgr():
-            self.before()
-            try:
-                yield lambda: self.callback()
-            except ExceptedException:
-                self.exception_call()
-            self.after()
-
-        self.mgr = context_mgr()
-
-    def test_normal_run(self):
-        thread = impl_eventlet.spawn_with(self.mgr, pool=eventlet)
-        thread.wait()
-        self.assertEqual(1, self.before.call_count)
-        self.assertEqual(1, self.callback.call_count)
-        self.assertEqual(1, self.after.call_count)
-        self.assertEqual(0, self.exception_call.call_count)
-
-    def test_excepted_exception(self):
-        self.callback.side_effect = ExceptedException
-        thread = impl_eventlet.spawn_with(self.mgr, pool=eventlet)
-        try:
-            thread.wait()
-        except ExceptedException:
-            pass
-        self.assertEqual(1, self.before.call_count)
-        self.assertEqual(1, self.callback.call_count)
-        self.assertEqual(1, self.after.call_count)
-        self.assertEqual(1, self.exception_call.call_count)
-
-    def test_unexcepted_exception(self):
-        self.callback.side_effect = Exception
-        thread = impl_eventlet.spawn_with(self.mgr, pool=eventlet)
-        try:
-            thread.wait()
-        except Exception:
-            pass
-        self.assertEqual(1, self.before.call_count)
-        self.assertEqual(1, self.callback.call_count)
-        self.assertEqual(0, self.after.call_count)
-        self.assertEqual(0, self.exception_call.call_count)
