@@ -20,6 +20,7 @@ import uuid
 
 import fixtures
 from oslo_serialization import jsonutils
+from oslo_utils import strutils
 from oslo_utils import timeutils
 from stevedore import dispatch
 from stevedore import extension
@@ -28,8 +29,8 @@ import yaml
 
 import oslo_messaging
 from oslo_messaging.notify import _impl_log
-from oslo_messaging.notify import _impl_messaging
 from oslo_messaging.notify import _impl_test
+from oslo_messaging.notify import messaging
 from oslo_messaging.notify import notifier as msg_notifier
 from oslo_messaging import serializer as msg_serializer
 from oslo_messaging.tests import utils as test_utils
@@ -144,7 +145,7 @@ class TestMessagingNotifier(test_utils.BaseTestCase):
         super(TestMessagingNotifier, self).setUp()
 
         self.logger = self.useFixture(_ReRaiseLoggedExceptionsFixture()).logger
-        self.stubs.Set(_impl_messaging, 'LOG', self.logger)
+        self.stubs.Set(messaging, 'LOG', self.logger)
         self.stubs.Set(msg_notifier, '_LOG', self.logger)
 
     @mock.patch('oslo_utils.timeutils.utcnow')
@@ -155,8 +156,9 @@ class TestMessagingNotifier(test_utils.BaseTestCase):
         if self.v2:
             drivers.append('messagingv2')
 
-        self.config(notification_driver=drivers,
-                    notification_topics=self.topics)
+        self.config(driver=drivers,
+                    topics=self.topics,
+                    group='oslo_messaging_notifications')
 
         transport = _FakeTransport(self.conf)
 
@@ -268,7 +270,8 @@ class TestLogNotifier(test_utils.BaseTestCase):
 
     @mock.patch('oslo_utils.timeutils.utcnow')
     def test_notifier(self, mock_utcnow):
-        self.config(notification_driver=['log'])
+        self.config(driver=['log'],
+                    group='oslo_messaging_notifications')
 
         transport = _FakeTransport(self.conf)
 
@@ -317,11 +320,27 @@ class TestLogNotifier(test_utils.BaseTestCase):
         msg = {'event_type': 'foo'}
         driver.notify(None, msg, "sample", None)
 
+    def test_mask_passwords(self):
+        # Ensure that passwords are masked with notifications
+        driver = _impl_log.LogDriver(None, None, None)
+        logger = mock.MagicMock()
+        logger.info = mock.MagicMock()
+        message = {'password': 'passw0rd', 'event_type': 'foo'}
+        json_str = jsonutils.dumps(message)
+        mask_str = strutils.mask_password(json_str)
+
+        with mock.patch.object(logging, 'getLogger') as gl:
+            gl.return_value = logger
+            driver.notify(None, message, 'info', 0)
+
+        logger.info.assert_called_once_with(mask_str)
+
 
 class TestRoutingNotifier(test_utils.BaseTestCase):
     def setUp(self):
         super(TestRoutingNotifier, self).setUp()
-        self.config(notification_driver=['routing'])
+        self.config(driver=['routing'],
+                    group='oslo_messaging_notifications')
 
         transport = _FakeTransport(self.conf)
         self.notifier = oslo_messaging.Notifier(transport)
@@ -343,13 +362,14 @@ class TestRoutingNotifier(test_utils.BaseTestCase):
         self.assertTrue(self.router._should_load_plugin(ext))
 
     def test_load_notifiers_no_config(self):
-        # default routing_notifier_config=""
+        # default routing_config=""
         self.router._load_notifiers()
         self.assertEqual({}, self.router.routing_groups)
         self.assertEqual(0, len(self.router.used_drivers))
 
     def test_load_notifiers_no_extensions(self):
-        self.config(routing_notifier_config="routing_notifier.yaml")
+        self.config(routing_config="routing_notifier.yaml",
+                    group='oslo_messaging_notifications')
         routing_config = r""
         config_file = mock.MagicMock()
         config_file.return_value = routing_config
@@ -365,7 +385,8 @@ class TestRoutingNotifier(test_utils.BaseTestCase):
         self.assertEqual({}, self.router.routing_groups)
 
     def test_load_notifiers_config(self):
-        self.config(routing_notifier_config="routing_notifier.yaml")
+        self.config(routing_config="routing_notifier.yaml",
+                    group='oslo_messaging_notifications')
         routing_config = r"""
 group_1:
    rpc : foo
@@ -395,7 +416,7 @@ group_1:
           - blah.zoo.*
           - zip
         """
-        groups = yaml.load(config)
+        groups = yaml.safe_load(config)
         group = groups['group_1']
 
         # No matching event ...
@@ -426,7 +447,7 @@ group_1:
           - info
           - error
         """
-        groups = yaml.load(config)
+        groups = yaml.safe_load(config)
         group = groups['group_1']
 
         # No matching priority
@@ -459,7 +480,7 @@ group_1:
       accepted_events:
           - foo.*
         """
-        groups = yaml.load(config)
+        groups = yaml.safe_load(config)
         group = groups['group_1']
 
         # Valid event, but no matching priority
@@ -502,7 +523,8 @@ group_1:
                                  sorted(pm.map.call_args[0][6]))
 
     def test_notify_filtered(self):
-        self.config(routing_notifier_config="routing_notifier.yaml")
+        self.config(routing_config="routing_notifier.yaml",
+                    group='oslo_messaging_notifications')
         routing_config = r"""
 group_1:
     rpc:
