@@ -15,6 +15,9 @@
 import stevedore
 import testtools
 
+import mock
+
+from oslo_messaging._executors import impl_thread
 try:
     from oslo_messaging import opts
 except ImportError:
@@ -29,15 +32,14 @@ class OptsTestCase(test_utils.BaseTestCase):
         super(OptsTestCase, self).setUp()
 
     def _test_list_opts(self, result):
-        self.assertEqual(6, len(result))
+        self.assertEqual(5, len(result))
 
         groups = [g for (g, l) in result]
         self.assertIn(None, groups)
-        self.assertIn('matchmaker_ring', groups)
         self.assertIn('matchmaker_redis', groups)
         self.assertIn('oslo_messaging_amqp', groups)
+        self.assertIn('oslo_messaging_notifications', groups)
         self.assertIn('oslo_messaging_rabbit', groups)
-        self.assertIn('oslo_messaging_qpid', groups)
 
         opt_names = [o.name for (g, l) in result for o in l]
         self.assertIn('rpc_backend', opt_names)
@@ -55,3 +57,8 @@ class OptsTestCase(test_utils.BaseTestCase):
 
         self.assertIsNotNone(result)
         self._test_list_opts(result)
+
+    def test_defaults(self):
+        impl_thread.ThreadExecutor(self.conf, mock.Mock(), mock.Mock())
+        opts.set_defaults(self.conf, executor_thread_pool_size=100)
+        self.assertEqual(100, self.conf.executor_thread_pool_size)
